# Install uvloop
try:
	import uvloop
	uvloop.install()
except ImportError:
	print("Couldn't install uvloop, falling back to the slower asyncio event loop")

import argparse
import asyncio
import pathlib
import os
import shutil
import logging

from aiohttp import web, log
from pippin.db.redis import RedisDB
from tortoise import Tortoise
from pippin.db.tortoise_config import DBConfig
from pippin.config import Config
from logging.handlers import TimedRotatingFileHandler, WatchedFileHandler
from pippin.network.rpc_client import RPCClient
from pippin.network.work_client import WorkClient
from pippin.server.pippin_server import PippinServer
from pippin.util.nano_util import NanoUtil
from pippin.util.utils import Utils

parser = argparse.ArgumentParser(description="Pippin Server")
parser.add_argument('--generate-config', action='store_true', help='Generate sample configuration file and exit', default=False)
options = parser.parse_args()

# Create sample file if not exists
config_dir = Utils.get_project_root()
sample_file = config_dir.joinpath(pathlib.PurePath('sample.config.yaml'))
real_file = config_dir.joinpath(pathlib.PurePath('config.yaml'))
if (not os.path.isfile(sample_file) and not os.path.isfile(real_file)) or options.generate_config:
    ref_file = pathlib.Path(__file__).parent.joinpath(pathlib.PurePath('sample.config.yaml'))
    shutil.copyfile(ref_file, sample_file)
    print(f"Sample configuration created at: {sample_file}")
    if options.generate_config:
        exit(0)

# Configuration
config = Config.instance()

# Set and patch nanopy
import nanopy
nanopy.account_prefix = 'ban_' if config.banano else 'nano_'
if config.banano:
    nanopy.standard_exponent = 29
    nanopy.work_difficulty = 'fffffe0000000000'

# Setup logger
if config.debug:
    logging.basicConfig(level=logging.DEBUG)
elif config.stdout:
    logging.basicConfig(level=logging.INFO)
else:
    root = logging.getLogger('aiohttp.server')
    logging.basicConfig(level=logging.INFO)
    handler = WatchedFileHandler(config.log_file)
    formatter = logging.Formatter("%(asctime)s;%(levelname)s;%(message)s", "%Y-%m-%d %H:%M:%S %z")
    handler.setFormatter(formatter)
    root.addHandler(handler)
    root.addHandler(TimedRotatingFileHandler(config.log_file, when="d", interval=1, backupCount=100))  

def main():
    loop = asyncio.get_event_loop()
    try:
        # Initialize database first
        log.server_logger.info("Initializing database")
        loop.run_until_complete(DBConfig().init_db())
        # Setup server
        server = PippinServer(config.host, config.port)
        # Check is remote node is alive
        try:
            is_alive = loop.run_until_complete(RPCClient.instance().is_alive())
        except Exception:
            log.server_logger.exception("Couldn't do is_alive RPC call")
            is_alive = False
        finally:
            if not is_alive:
                log.server_logger.error(f"Error: Could not connect to remote node at {Config.instance().node_url}")
                exit(1)
        # Start server
        log.server_logger.info(f"Pippin server starting at {config.host}:{config.port}")
        tasks = [
            server.start()
        ]
        # Check if DPoW should be started
        dpow_client = WorkClient.instance().dpow_client
        if dpow_client is not None:
            loop.run_until_complete(dpow_client.setup())
            tasks.append(dpow_client.loop())
        loop.run_until_complete(asyncio.wait(tasks))
        loop.run_forever()
    except Exception:
        log.server_logger.exception("Pippin exited with exception")
    except BaseException:
        pass
    finally:
        log.server_logger.info("Pipping is exiting")
        tasks = [
            RPCClient.close(),
            server.stop(),
            RedisDB.close(),
            WorkClient.close(),
            NanoUtil.close(),
            Tortoise.close_connections()
        ]
        loop.run_until_complete(asyncio.wait(tasks))
        loop.close()

if __name__ == "__main__":
    main()


import aiohttp
import ipaddress
import rapidjson as json
import socket
from pippin.config import Config
from typing import List

class RPCClient(object):
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls) -> 'RPCClient':
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls.node_url = Config.instance().node_url
            cls.session = aiohttp.ClientSession(json_serialize=json.dumps)
        return cls._instance


    @classmethod
    async def close(cls):
        if hasattr(cls, 'session') and cls.session is not None:
            await cls.session.close()
        if cls._instance is not None:
            cls._instance = None

    async def make_request(self, req_json: dict):
        async with self.session.post(self.node_url ,json=req_json, timeout=300) as resp:
            return await resp.json()

    async def account_balance(self, account: str) -> dict:
        account_balance = {
            'action': 'account_balance',
            'account': account
        }
        respjson = await self.make_request(account_balance)
        if 'balance' in respjson:
            return respjson
        return None

    async def account_info(self, account: str) -> dict:
        info_action = {
            'action': 'account_info',
            'account': account,
            'representative': True,
            'pending': True
        }
        respjson = await self.make_request(info_action)
        if 'error' not in respjson:
            return respjson
        elif respjson['error'].lower() == 'account not found':
            raise AccountNotFound(account)
        return None

    async def block_info(self, hash: str) -> dict:
        info_action = {
            'action': 'block_info',
            'hash': hash
        }
        respjson = await self.make_request(info_action)
        if 'error' not in respjson and 'contents' in respjson:
            respjson['contents'] = json.loads(respjson['contents'])
            return respjson
        elif respjson['error'].lower() == 'block not found':
            raise BlockNotFound(hash)
        return None

    async def process(self, block: dict, subtype: str = None) -> str:
        """RPC Process, return hash if successful"""
        process_action = {
            'action': 'process',
            'json_block': False,
            'block': json.dumps(block)
        }
        if subtype is not None:
            process_action['subtype'] = subtype
        return await self.make_request(process_action)

    async def accounts_balances(self, accounts: List[str]) -> dict:
        """Return accounts_balances for accounts"""
        balances_action = {
            'action': 'accounts_balances',
            'accounts': accounts
        }
        return await self.make_request(balances_action)

    async def accounts_frontiers(self, accounts: List[str]) -> dict:
        """Return accounts_frontiers for accounts"""
        frontiers_action = {
            'action': 'accounts_frontiers',
            'accounts': accounts
        }
        return await self.make_request(frontiers_action)

    async def accounts_pending(self, accounts: List[str]) -> dict:
        """Return accounts_pending for accounts"""
        pending_action = {
            'action': 'accounts_pending',
            'accounts': accounts
        }
        return await self.make_request(pending_action)

    async def pending(self, account: str, threshold: int) -> List[str]:
        """RPC Pending"""
        pending_action = {
            'action': 'pending',
            'account': account,
            'threshold': str(threshold)
        }
        resp = await self.make_request(pending_action)
        ret = []
        if 'blocks' in resp:
            if resp['blocks'] == '':
                return []
            for k,v in resp['blocks'].items():
                ret.append(k)
            return ret
        return None
            

    async def is_alive(self) -> bool:
        """Returns whether or not the remote node is alive"""
        # We could use 'block_count' or something simple
        # But some nodes may not whitelist actions like that,
        # so we test with an essential action like account_balance
        test_action = {
            'action': 'account_balance',
            'account': 'ban_1tipbotgges3ss8pso6xf76gsyqnb69uwcxcyhouym67z7ofefy1jz7kepoy' if Config.instance().banano else 'nano_3o7uzba8b9e1wqu5ziwpruteyrs3scyqr761x7ke6w1xctohxfh5du75qgaj'
        }
        respjson = await self.make_request(test_action)
        if 'error' in respjson or 'balance' in respjson:
            return True
        return False

class AccountNotFound(Exception):
    pass

class BlockNotFound(Exception):
    pass